mod timeline_day_divider;
mod timeline_item;
mod timeline_new_messages_divider;
mod timeline_placeholder;

use std::{
    collections::{HashMap, HashSet, VecDeque},
    pin::Pin,
    sync::Arc,
};

use futures::{lock::Mutex, pin_mut, Stream, StreamExt};
use gtk::{gio, glib, glib::clone, prelude::*, subclass::prelude::*};
use log::{error, warn};
use matrix_sdk::{
    deserialized_responses::SyncTimelineEvent,
    ruma::{EventId, OwnedEventId, OwnedTransactionId, TransactionId},
    Error as MatrixError,
};
pub use timeline_day_divider::TimelineDayDivider;
pub use timeline_item::{TimelineItem, TimelineItemExt, TimelineItemImpl};
pub use timeline_new_messages_divider::TimelineNewMessagesDivider;
pub use timeline_placeholder::{PlaceholderKind, TimelinePlaceholder};
use tokio::task::JoinHandle;

use super::{Event, Room, SupportedEvent, UnsupportedEvent};
use crate::{prelude::*, spawn_tokio};

#[derive(Debug, Default, Hash, Eq, PartialEq, Clone, Copy, glib::Enum)]
#[repr(u32)]
#[enum_type(name = "TimelineState")]
pub enum TimelineState {
    #[default]
    Initial,
    Loading,
    Ready,
    Error,
    Complete,
}

const MAX_BATCH_SIZE: usize = 20;
type BackwardStream =
    Pin<Box<dyn Stream<Item = Vec<matrix_sdk::Result<SyncTimelineEvent>>> + 'static + Send>>;

mod imp {
    use std::cell::{Cell, RefCell};

    use glib::object::WeakRef;
    use once_cell::sync::Lazy;

    use super::*;

    #[derive(Debug, Default)]
    pub struct Timeline {
        pub room: WeakRef<Room>,
        /// A store to keep track of related events that aren't known
        pub relates_to_events: RefCell<HashMap<OwnedEventId, Vec<OwnedEventId>>>,
        /// All events shown in the room history
        pub list: RefCell<VecDeque<TimelineItem>>,
        /// A Hashmap linking `EventId` to corresponding `Event`
        pub event_map: RefCell<HashMap<OwnedEventId, Event>>,
        /// Maps the temporary `EventId` of the pending Event to the real
        /// `EventId`
        pub pending_events: RefCell<HashMap<OwnedTransactionId, OwnedEventId>>,
        /// A Hashset of `EventId`s that where just redacted.
        pub redacted_events: RefCell<HashSet<OwnedEventId>>,
        pub state: Cell<TimelineState>,
        pub backward_stream: Arc<Mutex<Option<BackwardStream>>>,
        pub forward_handle: Arc<Mutex<Option<JoinHandle<()>>>>,
        /// Whether this timeline has a typing row.
        pub has_typing: Cell<bool>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for Timeline {
        const NAME: &'static str = "Timeline";
        type Type = super::Timeline;
        type Interfaces = (gio::ListModel,);
    }

    impl ObjectImpl for Timeline {
        fn properties() -> &'static [glib::ParamSpec] {
            static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
                vec![
                    glib::ParamSpecObject::builder::<Room>("room")
                        .construct_only()
                        .build(),
                    glib::ParamSpecBoolean::builder("empty").read_only().build(),
                    glib::ParamSpecEnum::builder("state", TimelineState::default())
                        .read_only()
                        .build(),
                ]
            });

            PROPERTIES.as_ref()
        }

        fn set_property(&self, _id: usize, value: &glib::Value, pspec: &glib::ParamSpec) {
            match pspec.name() {
                "room" => self.obj().set_room(value.get().unwrap()),
                _ => unimplemented!(),
            }
        }

        fn property(&self, _id: usize, pspec: &glib::ParamSpec) -> glib::Value {
            let obj = self.obj();

            match pspec.name() {
                "room" => obj.room().to_value(),
                "empty" => obj.is_empty().to_value(),
                "state" => obj.state().to_value(),
                _ => unimplemented!(),
            }
        }
    }

    impl ListModelImpl for Timeline {
        fn item_type(&self) -> glib::Type {
            TimelineItem::static_type()
        }

        fn n_items(&self) -> u32 {
            let mut len = self.list.borrow().len() as u32;

            if self.has_typing.get() {
                len += 1;
            }

            len
        }

        fn item(&self, position: u32) -> Option<glib::Object> {
            let position = position as usize;
            let list = self.list.borrow();

            if self.has_typing.get() && position == list.len() {
                return Some(TimelinePlaceholder::typing().upcast());
            }

            list.get(position).map(|o| o.clone().upcast())
        }
    }
}

glib::wrapper! {
    /// List of all loaded items in a room. Implements ListModel.
    ///
    /// There is no strict message ordering enforced by the Timeline; items
    /// will be appended/prepended to existing items in the order they are
    /// received by the server.
    ///
    /// This struct additionally keeps track of pending events that have yet to
    /// get an event ID assigned from the server.
    pub struct Timeline(ObjectSubclass<imp::Timeline>)
        @implements gio::ListModel;
}

// TODO:
// - [ ] Add and handle AnyEphemeralRoomEvent this includes read recipes
// - [ ] Add new message divider
impl Timeline {
    pub fn new(room: &Room) -> Self {
        glib::Object::builder().property("room", &room).build()
    }

    fn items_changed(&self, position: u32, removed: u32, added: u32) {
        let imp = self.imp();

        let last_new_message_date;

        // Insert date divider, this needs to happen before updating the position and
        // headers
        let added = {
            let position = position as usize;
            let added = added as usize;
            let mut list = imp.list.borrow_mut();

            let mut previous_timestamp = if position > 0 {
                list.get(position - 1)
                    .and_then(|item| item.downcast_ref::<Event>())
                    .and_then(|event| event.timestamp())
            } else {
                None
            };
            let mut dividers: Vec<(usize, TimelineDayDivider)> = vec![];
            let mut index = position;
            for current in list.range(position..position + added) {
                if let Some(current_timestamp) = current
                    .downcast_ref::<Event>()
                    .and_then(|event| event.timestamp())
                {
                    if Some(current_timestamp.ymd()) != previous_timestamp.as_ref().map(|t| t.ymd())
                    {
                        dividers.push((index, TimelineDayDivider::new(current_timestamp.clone())));
                        previous_timestamp = Some(current_timestamp);
                    }
                }
                index += 1;
            }

            let dividers_len = dividers.len();
            last_new_message_date = dividers.last().and_then(|(_, divider)| divider.date());
            for (added, (position, date)) in dividers.into_iter().enumerate() {
                list.insert(position + added, date.upcast());
            }

            (added + dividers_len) as u32
        };

        // Remove first day divider if a new one is added earlier with the same day
        let removed = {
            let mut list = imp.list.borrow_mut();
            if let Some(date) = list
                .get(position as usize + added as usize)
                .and_then(|item| item.downcast_ref::<TimelineDayDivider>())
                .and_then(|divider| divider.date())
            {
                if Some(date.ymd()) == last_new_message_date.as_ref().map(|date| date.ymd()) {
                    list.remove(position as usize + added as usize);
                    removed + 1
                } else {
                    removed
                }
            } else {
                removed
            }
        };

        // Update the header for events that are allowed to hide the header
        {
            let position = position as usize;
            let added = added as usize;
            let list = imp.list.borrow();

            let mut previous_sender = if position > 0 {
                list.get(position - 1)
                    .filter(|item| item.can_hide_header())
                    .and_then(|item| item.event_sender())
            } else {
                None
            };

            for current in list.range(position..position + added) {
                let current_sender = current.event_sender();

                if !current.can_hide_header() {
                    current.set_show_header(false);
                    previous_sender = None;
                } else if current_sender != previous_sender {
                    current.set_show_header(true);
                    previous_sender = current_sender;
                } else {
                    current.set_show_header(false);
                }
            }

            // Update the events after the new events
            for next in list.range((position + added)..) {
                // After an event with non hideable header the visibility for headers will be
                // correct
                if !next.can_hide_header() {
                    break;
                }

                // Once the sender changes we can be sure that the visibility for headers will
                // be correct
                if next.event_sender() != previous_sender {
                    next.set_show_header(true);
                    break;
                }

                // The `next` has the same sender as the `current`, therefore we don't show the
                // header and we need to check the event after `next`
                next.set_show_header(false);
            }
        }

        // Add relations to event
        {
            let list = imp.list.borrow();
            let mut relates_to_events = imp.relates_to_events.borrow_mut();
            let mut redacted_events = imp.redacted_events.borrow_mut();

            for event in list
                .range(position as usize..(position + added) as usize)
                .filter_map(|item| item.downcast_ref::<SupportedEvent>())
            {
                if let Some(relates_to) = relates_to_events.remove(&event.event_id()) {
                    let mut replacing_events = vec![];
                    let mut reactions = vec![];

                    for relation_event_id in relates_to {
                        let relation = self
                            .event_by_id(&relation_event_id)
                            .and_then(|event| event.downcast::<SupportedEvent>().ok())
                            .expect("Previously known event has disappeared");

                        if relation.is_replacing_event() {
                            replacing_events.push(relation);
                        } else if relation.is_reaction() {
                            reactions.push(relation);
                        }
                    }

                    if position != 0 || event.replacing_events().is_empty() {
                        event.append_replacing_events(replacing_events);
                    } else {
                        event.prepend_replacing_events(replacing_events);
                    }
                    event.add_reactions(reactions);

                    if event.redacted() {
                        redacted_events.insert(event.event_id());
                    }
                }
            }
        }

        self.notify("empty");

        self.upcast_ref::<gio::ListModel>()
            .items_changed(position, removed, added);

        self.remove_redacted_events();
    }

    fn remove_redacted_events(&self) {
        let imp = self.imp();
        let mut redacted_events_pos = Vec::with_capacity(imp.redacted_events.borrow().len());

        // Find redacted events in the list
        {
            let mut redacted_events = imp.redacted_events.borrow_mut();
            let list = imp.list.borrow();
            let mut i = list.len();
            let mut list = list.iter();

            while let Some(item) = list.next_back() {
                if let Some(event) = item.downcast_ref::<SupportedEvent>() {
                    if redacted_events.remove(&event.event_id()) {
                        redacted_events_pos.push(i - 1);
                    }
                    if redacted_events.is_empty() {
                        break;
                    }
                }
                i -= 1;
            }
        }

        let mut redacted_events_pos = &mut redacted_events_pos[..];
        // Sort positions to start from the end so positions are still valid
        // and to group calls to `items_changed`.
        redacted_events_pos.sort_unstable_by(|a, b| b.partial_cmp(a).unwrap());
        while let Some(pos) = redacted_events_pos.first() {
            let mut pos = pos.to_owned();
            let mut removed = 1;

            {
                let mut list = imp.list.borrow_mut();
                list.remove(pos);

                // Remove all consecutive previous redacted events.
                while let Some(next_pos) = redacted_events_pos.get(removed) {
                    if next_pos == &(pos - 1) {
                        pos -= 1;
                        removed += 1;
                        list.remove(pos);
                    } else {
                        break;
                    }
                }
                redacted_events_pos = &mut redacted_events_pos[removed..];

                // Remove the day divider before this event if it's not useful anymore.
                let day_divider_before = pos >= 1
                    && list
                        .get(pos - 1)
                        .filter(|item| item.is::<TimelineDayDivider>())
                        .is_some();
                let after = pos == list.len()
                    || list
                        .get(pos)
                        .filter(|item| item.is::<TimelineDayDivider>())
                        .is_some();

                if day_divider_before && after {
                    pos -= 1;
                    removed += 1;
                    list.remove(pos);
                }
            }

            self.upcast_ref::<gio::ListModel>()
                .items_changed(pos as u32, removed as u32, 0);
        }
    }

    fn add_hidden_events(&self, events: Vec<SupportedEvent>, at_front: bool) {
        let imp = self.imp();
        let mut relates_to_events = imp.relates_to_events.borrow_mut();

        // Group events by related event
        let mut new_relations: HashMap<_, Vec<_>> = HashMap::new();
        for event in events {
            if let Some(relates_to) = relates_to_events.remove(&event.event_id()) {
                let mut replacing_events = vec![];
                let mut reactions = vec![];

                for relation_event_id in relates_to {
                    let relation = self
                        .event_by_id(&relation_event_id)
                        .and_then(|event| event.downcast::<SupportedEvent>().ok())
                        .expect("Previously known event has disappeared");

                    if relation.is_replacing_event() {
                        replacing_events.push(relation);
                    } else if relation.is_reaction() {
                        reactions.push(relation);
                    }
                }

                if !at_front || event.replacing_events().is_empty() {
                    event.append_replacing_events(replacing_events);
                } else {
                    event.prepend_replacing_events(replacing_events);
                }
                event.add_reactions(reactions);
            }

            if let Some(relates_to_event) = event.related_event_id() {
                let relations = new_relations.entry(relates_to_event).or_default();
                relations.push(event);
            }
        }

        // Handle new relations
        let mut redacted_events = imp.redacted_events.borrow_mut();
        for (relates_to_event_id, new_relations) in new_relations {
            if let Some(relates_to_event) = self.event_by_id(&relates_to_event_id) {
                // Get the relations in relates_to_event otherwise they will be added in
                // items_changed and they might not be added at the right place.
                let mut relations: Vec<_> = relates_to_events
                    .remove(&relates_to_event_id)
                    .unwrap_or_default()
                    .into_iter()
                    .map(|event_id| {
                        self.event_by_id(&event_id)
                            .and_then(|event| event.downcast::<SupportedEvent>().ok())
                            .expect("Previously known event has disappeared")
                    })
                    .collect();

                if let Some(relates_to_event) = relates_to_event.downcast_ref::<SupportedEvent>() {
                    if at_front {
                        relations.splice(..0, new_relations);
                    } else {
                        relations.extend(new_relations);
                    }

                    let mut replacing_events = vec![];
                    let mut reactions = vec![];

                    for relation in relations {
                        if relation.is_replacing_event() {
                            replacing_events.push(relation);
                        } else if relation.is_reaction() {
                            reactions.push(relation);
                        }
                    }

                    if !at_front || relates_to_event.replacing_events().is_empty() {
                        relates_to_event.append_replacing_events(replacing_events);
                    } else {
                        relates_to_event.prepend_replacing_events(replacing_events);
                    }
                    relates_to_event.add_reactions(reactions);

                    if relates_to_event.redacted() {
                        redacted_events.insert(relates_to_event.event_id());
                    }
                }
            } else {
                // Store the new event if the `related_to` event isn't known, we will update the
                // `relates_to` once the `related_to` event is added to the list
                let relates_to_event = relates_to_events.entry(relates_to_event_id).or_default();

                let relations_ids: Vec<_> =
                    new_relations.iter().map(|event| event.event_id()).collect();
                if at_front {
                    relates_to_event.splice(..0, relations_ids);
                } else {
                    relates_to_event.extend(relations_ids);
                }
            }
        }
    }

    /// Load events at the start of the timeline.
    ///
    /// Returns `true` when no messages were added, but more can be loaded.
    pub async fn load(&self) -> bool {
        let imp = self.imp();

        if matches!(
            self.state(),
            TimelineState::Loading | TimelineState::Complete
        ) {
            return false;
        }

        self.set_state(TimelineState::Loading);
        self.add_loading_spinner();

        let matrix_room = self.room().matrix_room();
        let timeline_weak = self.downgrade().into();
        let backward_stream = imp.backward_stream.clone();
        let forward_handle = imp.forward_handle.clone();

        let handle: tokio::task::JoinHandle<matrix_sdk::Result<_>> = spawn_tokio!(async move {
            let mut backward_stream_guard = backward_stream.lock().await;
            let mut forward_handle_guard = forward_handle.lock().await;
            if backward_stream_guard.is_none() {
                let (forward_stream, backward_stream) = matrix_room.timeline().await?;

                let forward_handle = tokio::spawn(async move {
                    handle_forward_stream(timeline_weak, forward_stream).await;
                });

                if let Some(old_forward_handle) = forward_handle_guard.replace(forward_handle) {
                    old_forward_handle.abort();
                }

                backward_stream_guard
                    .replace(Box::pin(backward_stream.ready_chunks(MAX_BATCH_SIZE)));
            }

            Ok(backward_stream_guard.as_mut().unwrap().next().await)
        });

        match handle.await.unwrap() {
            Ok(Some(events)) => {
                let events: Vec<_> = events
                    .into_iter()
                    .filter_map(|event| match event {
                        Ok(event) => Some(event),
                        Err(error) => {
                            error!("Failed to load messages: {}", error);
                            None
                        }
                    })
                    .collect();

                let deser_events: Vec<_> = events
                    .iter()
                    .filter_map(|event| event.event.deserialize().ok())
                    .collect();
                let room = self.room();
                room.session()
                    .verification_list()
                    .handle_response_room(&room, deser_events.iter());
                room.update_latest_unread(deser_events.iter());

                let events: Vec<Event> = events
                    .into_iter()
                    .map(|event| Event::new(event, &room))
                    .collect();

                self.remove_loading_spinner();
                if events.is_empty() {
                    self.set_state(TimelineState::Error);
                    return false;
                }

                self.set_state(TimelineState::Ready);
                !self.prepend(events)
            }
            Ok(None) => {
                self.remove_loading_spinner();
                self.set_state(TimelineState::Complete);
                false
            }
            Err(error) => {
                error!("Failed to load timeline: {}", error);
                self.set_state(TimelineState::Error);
                self.remove_loading_spinner();
                false
            }
        }
    }

    async fn clear(&self) {
        let imp = self.imp();
        // Remove backward stream so that we create new streams
        let mut backward_stream = imp.backward_stream.lock().await;
        backward_stream.take();

        let mut forward_handle = imp.forward_handle.lock().await;
        if let Some(forward_handle) = forward_handle.take() {
            forward_handle.abort();
        }

        let length = imp.list.borrow().len();
        imp.relates_to_events.replace(HashMap::new());
        imp.list.replace(VecDeque::new());
        imp.event_map.replace(HashMap::new());
        imp.pending_events.replace(HashMap::new());
        imp.redacted_events.replace(HashSet::new());
        self.set_state(TimelineState::Initial);

        self.notify("empty");
        self.upcast_ref::<gio::ListModel>()
            .items_changed(0, length as u32, 0);
    }

    /// Append the new events
    pub fn append(&self, batch: Vec<Event>) {
        let imp = self.imp();

        if batch.is_empty() {
            return;
        }
        let mut added = batch.len();

        let index = {
            let index = {
                let mut list = imp.list.borrow_mut();
                // Extend the size of the list so that rust doesn't need to reallocate memory
                // multiple times
                list.reserve(batch.len());

                list.len()
            };

            let mut pending_events = imp.pending_events.borrow_mut();
            let mut hidden_events = vec![];

            for event in batch.into_iter() {
                if let Some(event_id) = event
                    .downcast_ref::<UnsupportedEvent>()
                    .and_then(|event| event.event_id())
                {
                    imp.event_map.borrow_mut().insert(event_id, event);
                    added -= 1;
                } else if let Ok(event) = event.downcast::<SupportedEvent>() {
                    let event_id = event.event_id();

                    if event.counts_as_unread() {
                        event.sender().set_latest_activity(
                            event
                                .origin_server_ts()
                                .map(|ts| ts.0.into())
                                .unwrap_or_default(),
                        );
                    }

                    if let Some(pending_id) = event
                        .transaction_id()
                        .and_then(|txn_id| pending_events.remove(&txn_id))
                    {
                        let mut event_map = imp.event_map.borrow_mut();

                        if let Some(pending_event) = event_map.remove(&pending_id) {
                            pending_event.set_pure_event(event.pure_event());
                            event_map.insert(event_id, pending_event);
                        };
                        added -= 1;
                    } else {
                        imp.event_map
                            .borrow_mut()
                            .insert(event_id, event.clone().upcast());
                        if event.is_hidden_event() {
                            hidden_events.push(event);
                            added -= 1;
                        } else {
                            imp.list.borrow_mut().push_back(event.upcast());
                        }
                    }
                } else {
                    added -= 1;
                }
            }

            self.add_hidden_events(hidden_events, false);

            index
        };

        self.items_changed(index as u32, 0, added as u32);
    }

    /// Append an event that wasn't yet fully sent and received via a sync
    pub fn append_pending(&self, txn_id: &TransactionId, event: SupportedEvent) {
        let imp = self.imp();

        imp.event_map
            .borrow_mut()
            .insert(event.event_id(), event.clone().upcast());

        imp.pending_events
            .borrow_mut()
            .insert(txn_id.to_owned(), event.event_id());

        let index = {
            let mut list = imp.list.borrow_mut();
            let index = list.len();

            if event.is_hidden_event() {
                self.add_hidden_events(vec![event], false);
                None
            } else {
                list.push_back(event.upcast());
                Some(index)
            }
        };

        if let Some(index) = index {
            self.items_changed(index as u32, 0, 1);
        }
    }

    /// Get the event with the given id from the local store.
    ///
    /// Use this method if you are sure the event has already been received.
    /// Otherwise use `fetch_event_by_id`.
    pub fn event_by_id(&self, event_id: &EventId) -> Option<Event> {
        self.imp().event_map.borrow().get(event_id).cloned()
    }

    /// Get the position of the event with the given id in this `Timeline`.
    pub fn find_event_position(&self, event_id: &EventId) -> Option<usize> {
        self.imp()
            .list
            .borrow()
            .iter()
            .enumerate()
            .find_map(|(pos, item)| {
                item.downcast_ref::<Event>()
                    .and_then(|event| event.event_id())
                    .filter(|item_event_id| item_event_id == event_id)
                    .map(|_| pos)
            })
    }

    /// Fetch the event with the given id.
    ///
    /// If the event can't be found locally, a request will be made to the
    /// homeserver.
    ///
    /// Use this method if you are not sure the event has already been received.
    /// Otherwise use `event_by_id`.
    pub async fn fetch_event_by_id(&self, event_id: &EventId) -> Result<Event, MatrixError> {
        if let Some(event) = self.event_by_id(event_id) {
            Ok(event)
        } else {
            let room = self.room();
            let matrix_room = room.matrix_room();
            let event_id_clone = event_id.to_owned();
            let handle =
                spawn_tokio!(async move { matrix_room.event(event_id_clone.as_ref()).await });
            match handle.await.unwrap() {
                Ok(room_event) => Ok(Event::new(room_event.into(), &room)),
                Err(error) => {
                    // TODO: Retry on connection error?
                    warn!("Could not fetch event {}: {}", event_id, error);
                    Err(error)
                }
            }
        }
    }

    /// Prepends a batch of events.
    ///
    /// Returns `true` if new shown events where added to the timeline.
    pub fn prepend(&self, batch: Vec<Event>) -> bool {
        let imp = self.imp();
        let mut added = batch.len();

        {
            let mut hidden_events: Vec<_> = vec![];
            // Extend the size of the list so that rust doesn't need to reallocate memory
            // multiple times
            imp.list.borrow_mut().reserve(added);

            for event in batch {
                if let Some(event_id) = event
                    .downcast_ref::<UnsupportedEvent>()
                    .and_then(|event| event.event_id())
                {
                    imp.event_map.borrow_mut().insert(event_id, event);
                    added -= 1;
                } else if let Ok(event) = event.downcast::<SupportedEvent>() {
                    if event.counts_as_unread() {
                        event.sender().set_latest_activity(
                            event
                                .origin_server_ts()
                                .map(|ts| ts.0.into())
                                .unwrap_or_default(),
                        );
                    }

                    imp.event_map
                        .borrow_mut()
                        .insert(event.event_id(), event.clone().upcast());

                    if event.is_hidden_event() {
                        hidden_events.push(event);
                        added -= 1;
                    } else {
                        imp.list.borrow_mut().push_front(event.upcast());
                    }
                } else {
                    added -= 1;
                }
            }
            self.add_hidden_events(hidden_events, true);
        }

        self.items_changed(0, 0, added as u32);
        added > 0
    }

    /// Set the room containing this timeline.
    fn set_room(&self, room: Option<Room>) {
        self.imp().room.set(room.as_ref());

        if let Some(room) = room {
            room.typing_list().connect_items_changed(
                clone!(@weak self as obj => move |list, _, _, _| {
                    if !list.is_empty() {
                        obj.add_typing_row();
                    }
                }),
            );
        }
    }

    /// The room containing this timeline.
    pub fn room(&self) -> Room {
        self.imp().room.upgrade().unwrap()
    }

    fn set_state(&self, state: TimelineState) {
        let imp = self.imp();

        if state == self.state() {
            return;
        }

        imp.state.set(state);

        self.notify("state");
    }

    /// The state of the timeline.
    pub fn state(&self) -> TimelineState {
        self.imp().state.get()
    }

    /// Whether the timeline is empty.
    pub fn is_empty(&self) -> bool {
        let imp = self.imp();
        imp.list.borrow().is_empty()
            || (imp.list.borrow().len() == 1 && self.state() == TimelineState::Loading)
    }

    fn add_loading_spinner(&self) {
        self.imp()
            .list
            .borrow_mut()
            .push_front(TimelinePlaceholder::spinner().upcast());
        self.upcast_ref::<gio::ListModel>().items_changed(0, 0, 1);
    }

    fn remove_loading_spinner(&self) {
        self.imp().list.borrow_mut().pop_front();
        self.upcast_ref::<gio::ListModel>().items_changed(0, 1, 0);
    }

    fn has_typing_row(&self) -> bool {
        self.imp().has_typing.get()
    }

    fn add_typing_row(&self) {
        if self.has_typing_row() {
            return;
        }

        let pos = self.n_items();
        self.imp().has_typing.set(true);
        self.upcast_ref::<gio::ListModel>().items_changed(pos, 0, 1);
    }

    pub fn remove_empty_typing_row(&self) {
        if !self.has_typing_row() || !self.room().typing_list().is_empty() {
            return;
        }

        let pos = self.n_items() - 1;
        self.imp().has_typing.set(false);
        self.upcast_ref::<gio::ListModel>().items_changed(pos, 1, 0);
    }

    /// Remove the given event from the timeline.
    ///
    /// This should be used when the source of an `Event` changes and it should
    /// be hidden now.
    pub fn remove_event(&self, event_id: &EventId) {
        if let Some(index) = self.find_event_position(event_id) {
            self.imp().list.borrow_mut().remove(index);

            self.items_changed(index as u32, 1, 0);
        }
    }

    /// Replace the supported event with the given ID with the given unsupported
    /// event in this timeline.
    ///
    /// This should be used when the source of a `SupportedEvent` changes, and
    /// the new source fails to deserialize.
    pub fn replace_supported_event(&self, event_id: OwnedEventId, event: UnsupportedEvent) {
        self.remove_event(&event_id);

        self.imp()
            .event_map
            .borrow_mut()
            .insert(event_id, event.upcast());
    }
}

async fn handle_forward_stream(
    timeline: glib::SendWeakRef<Timeline>,
    stream: impl Stream<Item = SyncTimelineEvent>,
) {
    let stream = stream.ready_chunks(MAX_BATCH_SIZE);
    pin_mut!(stream);

    while let Some(events) = stream.next().await {
        let timeline = timeline.clone();
        let (sender, receiver) = futures::channel::oneshot::channel();
        let ctx = glib::MainContext::default();
        ctx.spawn(async move {
            let result = if let Some(timeline) = timeline.upgrade() {
                let events: Vec<_> = events
                    .into_iter()
                    .map(|event| Event::new(event, &timeline.room()))
                    .collect();

                timeline.append(events);

                true
            } else {
                false
            };
            sender.send(result).unwrap();
        });

        if !receiver.await.unwrap() {
            break;
        }
    }

    let ctx = glib::MainContext::default();
    ctx.spawn(async move {
        crate::spawn!(async move {
            if let Some(timeline) = timeline.upgrade() {
                timeline.clear().await;
            };
        });
    });
}
